"""Open URLs or files"""
import logging
from shutil import which
from .scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "xdg_open"
        if which("xdg-open") is None:
            raise SystemError("System command 'xdg-open' not found")

    def start_control(self, topic, data):
        """Control system"""
        logger.info("/usr/bin/xdg-open %s", data)
        syscommand(f"/usr/bin/xdg-open {data}")



"""Monitors for shutdown/sleep events"""

import threading
import signal
import logging
import traceback

from .modules.scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class MonitorSuspend:
    """Monitor DBUS for Suspend and Shutdown events"""

    def __init__(self, callback):
        self.use = False
        self._requirements()
        if self.lib["dbus"] is None:
            logger.error("Can't use DBus for system monitor")
            return
        try:
            dbus_loop = self.lib["dbus-loop"].mainloop.glib.DBusGMainLoop(
                set_as_default=True
            )  # integrate into main loob
            bus = self.lib["dbus"].SystemBus(
                mainloop=dbus_loop
            )  # connect to dbus system wide
            bus.add_signal_receiver(  # defince the signal to listen to
                callback,  # name of callback function
                "PrepareForSleep",  # singal name
                "org.freedesktop.login1.Manager",  # interface
                "org.freedesktop.login1",  # bus name
            )
            bus.add_signal_receiver(  # defince the signal to listen to
                callback,  # name of callback function
                "PrepareForShutdown",  # singal name
                "org.freedesktop.login1.Manager",  # interface
                "org.freedesktop.login1",  # bus name
            )
            self.loop = self.lib["glib"].repository.GLib.MainLoop()
            self.timer1 = threading.Thread(target=self.loop.run)
            self.use = True
        except Exception as err:
            logger.error(
                "Can't use DBus: %s, %s",
                err,
                traceback.format_exc(),
            )

    def _requirements(self):
        dbus = import_install_package("dbus-python", ">=1.3.2", "dbus")
        if dbus is None:
            self.lib = {"dbus": dbus}
            return
        self.lib = {
            "dbus": dbus,
            "dbus-loop": import_install_package(
                "dbus-python", ">=1.3.2", "dbus.mainloop.glib"
            ),
            "cairo": import_install_package("pycairo", ">=1.24.0", "cairo"),
            "glib": import_install_package(
                "PyGObject", ">=3.44.0", "gi.repository.GLib"
            ),
        }

    def start(self):
        """Start the timer of the thread"""
        if self.use:
            self.timer1.start()

    def stop(self):
        """Stop the timer of the thread"""
        if self.use:
            self.loop.quit()
            self.timer1.join()


class GracefulKiller:
    """Monitor Signal Termination"""

    def __init__(self, callback):
        """Checks for termination signals"""
        self.kill_now = False
        self.callback = callback
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, signum, frame):
        """Stop the app gracefully"""
        logger.info("stopped gracefully")
        self.callback(1)
        self.kill_now = True



"""Gets the idle time of the system"""
import sys
import logging
import importlib
from .scripts.helpers import import_install_package, syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Idle"
        self._requirements()
        if self.lib["dbus_idle"] is None:
            raise SystemError("Python package 'dbus_idle' can't be installed")

    def _requirements(self):
        self.lib = {
            "dbus_idle": import_install_package(
                "dbus-idle", ">=2023.12.0", "dbus_idle"
            ),
        }

    def get_info(self):
        """Gather information from the system"""
        try:
            idle_ms = self.lib["dbus_idle"].IdleMonitor().get_dbus_idle()
            idle_sec = round(idle_ms / 1000, 0)
            return idle_sec
        except NotImplementedError:
            logger.error("Unsupported version of dbus-idle found. Try to update...")
            args = [
                sys.executable,
                "-m",
                "pip",
                "install",
                "-U",
                "--quiet",
                "dbus-idle",
            ]
            syscommand(args, timeout=None)
            importlib.reload(self.lib["dbus_idle"])
            return 0
        except Exception as err:
            logger.error(err)
            return 0

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Idle": {
                "type": "sensor",
                "icon": "mdi:timer-sand",
                "unit": "s",
                "state_class": "total_increasing",
                "device_class": "duration",
            },
        }



"""Checks if the microphone is used"""
import glob
import json
import re
from .scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Microphone used"
        self.lnxlink = lnxlink

        _, _, returncode = syscommand(
            "which pactl && pactl -f json list short source-outputs", ignore_errors=True
        )
        self.use_pactl = False
        if returncode == 0:
            self.use_pactl = True

    def get_info(self):
        """Gather information from the system"""
        if self.use_pactl:
            stdout, _, _ = syscommand(
                "pactl -f json list short source-outputs",
            )
            # Replace 0,00 values with 0.00
            stdout = re.sub(r"(\s*[+-]?[0-9]+),([0-9]+\s*)", r"\1.\2", stdout)
            data = json.loads(stdout)
            if len(data) > 0:
                return "ON"
            return "OFF"
        mics = glob.glob("/proc/asound/**/*c/sub*/status", recursive=True)
        for mic in mics:
            with open(mic, encoding="UTF-8") as mic_content:
                mic_status = mic_content.read().strip().lower()
                if mic_status != "closed":
                    return "ON"
        return "OFF"

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Microphone used": {
                "type": "binary_sensor",
                "icon": "mdi:microphone",
            },
        }



"""Gets the battery information of connected devices"""
from shutil import which
import jc
from .scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Battery"
        self.lnxlink = lnxlink
        if which("upower") is None:
            raise SystemError("System command 'upower' not found")
        self.devices = self._get_devices()

    def exposed_controls(self):
        """Exposes to home assistant"""
        discovery_info = {}
        for device in self.devices:
            att_temp = f"{{{{ value_json.get('{device}', {{}}).get('attributes', {{}}) | tojson }}}}"
            discovery_info[f"Battery {device}"] = {
                "type": "sensor",
                "icon": "mdi:battery",
                "unit": "%",
                "device_class": "battery",
                "value_template": f"{{{{ value_json.get('{device}', {{}}).get('percent') }}}}",
                "attributes_template": att_temp,
                "enabled": True,
            }
        return discovery_info

    def get_info(self):
        """Gather information from the system"""
        devices = self._get_devices()
        new_devices = set(devices) - set(self.devices)
        disconnected_devices = set(self.devices) - set(devices)
        for device_name in disconnected_devices:
            devices[device_name] = self.devices[device_name]
            self.devices[device_name]["percent"] = None
        self.devices = devices
        if len(new_devices) > 0:
            self.lnxlink.setup_discovery()

        return devices

    def _get_devices(self):
        stdout, _, _ = syscommand("upower --dump")
        upower_json = jc.parse("upower", stdout)

        devices = {}
        for device in upower_json:
            if "detail" in device:
                if "percentage" in device["detail"]:
                    native_path = device.get("native_path", "").split("/")[-1]
                    name = " ".join(
                        [
                            device.get("vendor", ""),
                            device.get("model", ""),
                            device.get("serial", "").replace(":", ""),
                        ]
                    ).strip()
                    if name == "":
                        name = native_path
                    if name != "":
                        devices[name] = {
                            "percent": device["detail"]["percentage"],
                            "attributes": {
                                "vendor": device.get("vendor", ""),
                                "model": device.get("model", ""),
                                "serial": device.get("serial", ""),
                                "native_path": native_path,
                                "rechargeable": device["detail"].get(
                                    "rechargeable", ""
                                ),
                            },
                        }
        return devices



"""Gets WiFi information"""
import os
import re
from .scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "WiFi"

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "WiFi": {
                "type": "sensor",
                "icon": "mdi:wifi",
                "unit": "%",
                "entity_category": "diagnostic",
                "state_class": "measurement",
                "value_template": "{{ value_json.signal }}",
                "attributes_template": "{{ value_json.attributes | tojson }}",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        interface = ""
        ssid = ""
        mac = ""
        signal = None
        if os.path.exists("/proc/net/wireless"):
            wireless_info, _, _ = syscommand("cat /proc/net/wireless")
            match = re.findall(r"\s+(\S+):\s\S+\s+\S+\s+(\S+)", wireless_info)
            if match:
                interface = match[0][0]
                rssi = float(match[0][1])
                signal = min(2 * (100 + rssi), 100)
                ssid, _, _ = syscommand("iwgetid -r")
                mac, _, _ = syscommand("iwgetid -ra")

        return {
            "signal": signal,
            "attributes": {
                "Interface": interface,
                "SSID": ssid,
                "MAC": mac,
            },
        }



"""Restarts the system"""
import logging
from .scripts.helpers import syscommand

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Restart"
        self.lnxlink = lnxlink

    def start_control(self, topic, data):
        """Control system"""
        self.lnxlink.temp_connection_callback(True)
        _, _, returncode = syscommand("systemctl reboot")
        if returncode != 0:
            _, _, returncode = syscommand("shutdown -r now")
            if returncode != 0:
                self.lnxlink.temp_connection_callback(False)
                logger.error("Can't shutdown the computer")

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "restart": {
                "type": "button",
                "icon": "mdi:restart",
            }
        }



"""Shows notifications"""
import logging
import requests
from .scripts.helpers import import_install_package

logger = logging.getLogger("lnxlink")


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Notify OSD"
        self._requirements()
        self.lib["dbus"].mainloop.glib.DBusGMainLoop(set_as_default=True)

    def _requirements(self):
        self.lib = {
            "notify2": import_install_package("notify2", ">=0.3.1"),
            "dbus": import_install_package(
                "dbus-python", ">=1.3.2", "dbus.mainloop.glib"
            ),
        }

    def start_control(self, topic, data):
        """Control system"""
        icon_url = data.get("iconUrl", "")
        icon = "notification-message-im"
        try:
            if "http" in icon_url:
                icon_ext = icon_url.split(".")[-1]
                img_data = requests.get(icon_url, timeout=3).content
                icon = f"/tmp/lnxlink_notification.{icon_ext}"
                with open(icon, "wb") as handler:
                    handler.write(img_data)
            elif icon_url != "":
                icon = icon_url
        except Exception as err:
            logger.error("Error downloading notification image: %s", err)

        # notify2
        self.lib["notify2"].init("lnxlink")
        self.lib["notify2"].Notification(data["title"], data["message"], icon).show()



"""Gets CPU usage information"""
import psutil


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "CPU Usage"
        self.lnxlink = lnxlink

    def get_info(self):
        """Gather information from the system"""
        return psutil.cpu_percent()

    def exposed_controls(self):
        """Exposes to home assistant"""
        update_interval = self.lnxlink.config.get("update_interval", 5)
        return {
            "CPU Usage": {
                "type": "sensor",
                "icon": "mdi:speedometer",
                "unit": "%",
                "state_class": "measurement",
                "expire_after": update_interval * 5,
            },
        }



"""Checks if the webcam is used"""
from .scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Camera used"

    def get_info(self):
        """Gather information from the system"""
        _, _, returncode = syscommand("fuser /dev/video*", ignore_errors=True)
        cam_used = False
        if returncode == 0:
            cam_used = True

        return cam_used

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Camera used": {
                "type": "binary_sensor",
                "icon": "mdi:webcam",
            },
        }



"""Checks for system updates"""
import time
from .scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "System Updates"
        self.last_time = 0
        self.update_interval = 360  # Check for updates every 6 minutes
        self.update_available = False

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "System Updates": {
                "type": "binary_sensor",
                "icon": "mdi:package-variant",
                "entity_category": "diagnostic",
            },
        }

    def get_info(self):
        """Gather information from the system"""
        packages = [
            {
                "command": "apt list --upgradable | wc -l",
                "logic": "> 2",
            },
            {
                "command": "yum -q updateinfo list updates | wc -l",
                "logic": "> 2",
            },
        ]

        cur_time = time.time()
        if cur_time - self.last_time > self.update_interval:
            self.last_time = cur_time
            for package in packages:
                stdout, _, _ = syscommand(package["command"])
                self.update_available = eval(f"{stdout}{package['logic']}")
                if self.update_available:
                    return True
        return self.update_available



"""A collection of helper functions"""
import sys
import logging
import subprocess

logger = logging.getLogger("lnxlink")


def syscommand(command, ignore_errors=False, timeout=3, background=False):
    """Global subprocess command"""
    if isinstance(command, list):
        command = " ".join(command)
    if background:
        subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        return "", "", 0
    result = subprocess.run(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=False,
        timeout=timeout,
    )
    stderr = result.stderr.decode("UTF-8").strip()
    stdout = result.stdout.decode("UTF-8").strip()
    returncode = result.returncode
    if returncode != 0 and ignore_errors is False:
        logger.error("Error with command: %s (%s)", command, stderr)
    return stdout, stderr, returncode


def import_install_package(package, version="", syspackage=None):
    """Imports a system package and if it doesn't exist, it gets installed"""
    if syspackage is None:
        syspackage = package
    try:
        return __import__(syspackage)
    except ImportError:
        package_version = f"'{package}{version}'"
        logger.info("Package %s is not installed, installing now...", package_version)
        args = [sys.executable, "-m", "pip", "install", "--quiet", package_version]
        _, _, returncode = syscommand(args, ignore_errors=True, timeout=None)
        if returncode != 0:
            logger.error("Can't install package %s", package)
            return None
    return __import__(syspackage)



"""Run a terminal command"""
from .scripts.helpers import syscommand


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "bash"
        self.lnxlink = lnxlink

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Bash_Command": {
                "type": "text",
                "icon": "mdi:bash",
            }
        }

    def start_control(self, topic, data):
        """Control system"""
        stdout, _, _ = syscommand(data, timeout=120)
        return stdout



"""Shows an image of the desktop as a camera entity"""
import base64
from .scripts.helpers import import_install_package


class Addon:
    """Addon module"""

    def __init__(self, lnxlink):
        """Setup addon"""
        self.name = "Screenshot"
        self.run = False
        self._requirements()

    def _requirements(self):
        self.lib = {
            "cv2": import_install_package("opencv-python", ">=4.7.0.68", "cv2"),
            "mss": import_install_package("mss", ">=7.0.1"),
            "np": import_install_package("numpy", ">=1.24.0"),
        }

    def get_camera_frame(self):
        """Convert screen image to Base64 text"""
        if self.run:
            with self.lib["mss"].mss() as sct:
                sct_img = sct.grab(sct.monitors[1])
                frame = self.lib["np"].array(sct_img)
            _, buffer = self.lib["cv2"].imencode(".jpg", frame)
            frame = base64.b64encode(buffer)
            return frame
        return None

    def get_info(self):
        """Gather information from the system"""
        return self.run

    def exposed_controls(self):
        """Exposes to home assistant"""
        return {
            "Screenshot": {
                "type": "switch",
                "icon": "mdi:monitor-screenshot",
                "entity_category": "config",
            },
            "Screenshot feed": {
                "type": "camera",
                "method": self.get_camera_frame,
                "encoding": "b64",
            },
        }

    def start_control(self, topic, data):
        """Control system"""
        if data.lower() == "off":
            self.run = False
        elif data.lower() == "on":
            self.run = True
